package ch.unige.solidify.util;

import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.text.ParseException;
import java.util.Arrays;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.boot.web.client.RestTemplateCustomizer;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.web.client.RequestCallback;
import org.springframework.web.client.ResponseExtractor;
import org.springframework.web.client.RestTemplate;

import com.nimbusds.jwt.SignedJWT;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.interceptor.PlusEncoderInterceptor;
import ch.unige.solidify.security.TokenUsage;

public abstract class AbstractRestClientTool {
  private static final Logger log = LoggerFactory.getLogger(AbstractRestClientTool.class);

  private static class UTF8MessageConverterCustomizer implements RestTemplateCustomizer {
    @Override
    public void customize(RestTemplate restTemplate) {
      final List<HttpMessageConverter<?>> converters = restTemplate.getMessageConverters();
      converters.removeIf(httpMessageConverter -> httpMessageConverter instanceof StringHttpMessageConverter);
      converters.add(0, new StringHttpMessageConverter(StandardCharsets.UTF_8));
    }
  }

  public void downloadContent(URI uri, Path path, TokenUsage tokenUsage) {
    ResponseExtractor<Void> responseExtractor = this.buildExtractor(uri, path);
    final RestTemplate restTemplate;
    if (tokenUsage == TokenUsage.WITH_TOKEN) {
      restTemplate =  this.getRestTemplateBuilder().build();
    } else if (tokenUsage == TokenUsage.WITHOUT_TOKEN) {
      restTemplate = new RestTemplateBuilder().build();
    } else {
      throw new SolidifyRuntimeException("Unknown token usage value: " + tokenUsage);
    }
    this.launchDownload(uri, responseExtractor, restTemplate);
  }


  private void launchDownload(URI uri, ResponseExtractor<Void> responseExtractor, RestTemplate restTemplate) {
    // Optional Accept header
    final RequestCallback requestCallback = request -> request.getHeaders()
            .setAccept(Arrays.asList(MediaType.APPLICATION_OCTET_STREAM, MediaType.ALL));
    restTemplate.execute(uri, HttpMethod.GET, requestCallback, responseExtractor);
  }

  private ResponseExtractor<Void> buildExtractor(URI uri, Path path) {
    return response -> {
      final byte[] data = new byte[SolidifyConstants.BUFFER_SIZE];
      try (final InputStream inputStream = response.getBody();
           final FileOutputStream fileOutputStream = new FileOutputStream(path.toFile());
           final BufferedOutputStream bufferedOutputStream = new BufferedOutputStream(fileOutputStream, SolidifyConstants.BUFFER_SIZE)) {
        int nRead;
        while ((nRead = inputStream.read(data, 0, data.length)) != -1) {
          bufferedOutputStream.write(data, 0, nRead);
        }
      } catch (final IOException e) {
        throw new SolidifyRuntimeException("Error downloading " + uri.toString() + " to " + path.toString() + uri, e);
      }
      return null;
    };
  }

  public RestTemplate getClient() {
    return this.getRestTemplateBuilder().build();
  }

  public RestTemplate getClientWithPlusEncoderInterceptor() {
    return this.getRestTemplateBuilder().additionalInterceptors(new PlusEncoderInterceptor()).build();
  }

  private RestTemplateBuilder getRestTemplateBuilder() {
    RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
    restTemplateBuilder = this.addToken(restTemplateBuilder);
    restTemplateBuilder = this.addUTF8Converter(restTemplateBuilder);
    return restTemplateBuilder;
  }

  private RestTemplateBuilder addUTF8Converter(RestTemplateBuilder restTemplateBuilder) {
    return restTemplateBuilder.additionalCustomizers(new UTF8MessageConverterCustomizer());
  }

  protected abstract RestTemplateBuilder addToken(RestTemplateBuilder restTemplateBuilder);

  protected String parseToken(String token) {
    try {
      return SignedJWT.parse(token).getJWTClaimsSet().toString();
    } catch (ParseException e) {
      return e.getMessage();
    }
  }

  private String getUrl(URL url) {
    String rootUrl = url.getProtocol() + "://" + url.getHost();
    if (url.getPort() > 0) {
      rootUrl += ":" + url.getPort();
    }
    return rootUrl + url.getPath();
  }

}
