package ch.unige.solidify.service;

import static ch.unige.solidify.util.StringTool.isNullOrEmpty;

import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.BindingResult;
import org.springframework.validation.Validator;

import ch.unige.solidify.exception.SolidifyValidationException;
import ch.unige.solidify.repository.SolidifyRepository;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.Searchable;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.specification.SolidifySpecification;
import ch.unige.solidify.util.SearchCriteria;
import ch.unige.solidify.util.SearchOperation;
import ch.unige.solidify.validation.ValidationError;

public abstract class ResourceService<T extends Resource<T>> extends AbstractResourceService {

  @Autowired
  private Validator validator;

  @Autowired
  protected SolidifyRepository<T> itemRepository;

  /**
   * Find all subresources of a given type, with a specific parentId
   *
   * @param parentId    the id of the parent
   * @param parentClass the class of the resource
   * @param childClass  the class of the subresource
   * @param pageable    the pagination information
   * @param <V>         the type of the subresource as a parameter type
   * @return
   */
  public <V> Page<V> findByParentId(String parentId, Class<?> parentClass, Class<V> childClass,
          Pageable pageable) {
    return this.itemRepository.findByParentId(parentId, parentClass, childClass, pageable);
  }

  public <V> V findByParentIdAndChildId(String parentId, Class<?> parentClass, Class<V> childClass,
          String id) {
    return this.itemRepository.findByParentIdAndChildId(parentId, parentClass, childClass, id);
  }

  protected T localSave(T item) {
    return this.itemRepository.save(item);
  }

  protected T localFindOne(String id) {
    return this.itemRepository.findById(id)
            .orElseThrow(() -> new NoSuchElementException("No resource with id " + id));
  }

  protected boolean localExistsById(String resId) {
    return this.itemRepository.existsById(resId);
  }

  protected Page<T> localFindAll(Pageable pageable) {
    return this.itemRepository.findAll(pageable);
  }

  protected Page<T> localFindAll(Specification<T> spec, Pageable pageable) {
    return this.itemRepository.findAll(spec, pageable);
  }

  public Page<T> findAll(Pageable pageable) {
    return this.localFindAll(pageable);
  }

  public Page<T> findAll(Specification<T> spec, Pageable pageable) {
    final Specification<T> completedSpec = this.addSpecificationFilter(spec);
    return this.localFindAll(completedSpec, pageable);
  }

  public T findOne(String resId) {
    return this.localFindOne(resId);
  }

  public Page<T> findBySearchCriteria(Searchable<T> resource, String orOpredicateFlag, List<SearchCriteria> criterias, Pageable pageable) {
    return this.localFindBySearchCriteria(resource, orOpredicateFlag, criterias, pageable);
  }

  public Page<T> findBySubResourceSearchCriteria(Searchable<T> subresource, String orOpredicateFlag, List<SearchCriteria> criterias,
          SearchCriteria parent, Pageable pageable) {
    return this.localFindBySubResourceSearchCriteria(subresource, orOpredicateFlag, criterias, parent, pageable);
  }

  protected Page<T> localFindBySearchCriteria(Searchable<T> resource, String orOpredicateFlag, List<SearchCriteria> criterias,
          Pageable pageable) {
    if (criterias == null || criterias.isEmpty()) {
      return null;
    }
    final Specification<T> spec = this.localFindBySearchCriteria(resource, orOpredicateFlag, criterias);
    return this.localFindAll(spec, pageable);
  }

  protected Page<T> localFindBySubResourceSearchCriteria(Searchable<T> resource, String orOpredicateFlag, List<SearchCriteria> criterias,
          SearchCriteria parent, Pageable pageable) {

    if (criterias == null || criterias.isEmpty()) {
      return null;
    }
    Specification<T> spec = this.localFindBySearchCriteria(resource, orOpredicateFlag, criterias);

    spec = Specification.where(spec).and(resource.getSearchSpecification(parent));
    return this.localFindAll(spec, pageable);
  }

  protected void localDelete(String id) {
    this.itemRepository.deleteById(id);
  }

  protected void localDeleteAll(Iterable<? extends T> resources) {
    this.itemRepository.deleteAll(resources);
  }

  public void delete(String id) {
    this.localDelete(id);
  }

  public void deleteAll(Iterable<? extends T> resources) {
    this.localDeleteAll(resources);
  }

  public boolean existsById(String resId) {
    return this.localExistsById(resId);
  }

  public <S extends Resource<S>> S getSubResource(String subResourceId, Class<S> subResourceType) {
    final RemoteResourceService<S> service = this.getResourceService(subResourceType);
    return service.findOneWithCache(subResourceId);
  }

  public <S extends Resource<S>, U extends SubResourceContainer> Collection<S> getSubResources(U parentItem, Class<S> subResourceType,
          Pageable pageable) {

    Page<String> subResourceIds = parentItem.getSubResourceIds(subResourceType, pageable);

    final List<S> items = new ArrayList<>();

    if (subResourceIds != null) {
      for (final String subResourceId : subResourceIds) {
        if (!isNullOrEmpty(subResourceId)) {
          final S item = this.getSubResource(subResourceId, subResourceType);
          if (item != null) {
            items.add(item);
          }
        }
      }
      Page page = new PageImpl(items, pageable, subResourceIds.getTotalElements());
      return new Collection<>(page, pageable);
    } else {
      return new Collection<>();
    }
  }

  protected <S extends Resource<S>> RemoteResourceService<S> getResourceService(Class<S> type) {
    throw new UnsupportedOperationException("Missing resource service for type " + type.getName());
  }

  /**
   * Override to add criteria to the Specification. May be useful to add criteria based on permissions
   * fetched from Authentication
   *
   * @param spec
   * @return
   */
  protected Specification<T> addSpecificationFilter(Specification<T> spec) {
    return spec;
  }

  private Specification<T> localFindBySearchCriteria(Searchable<T> resource, String orOpredicateFlag, List<SearchCriteria> criterias) {
    Specification<T> spec = resource.getSearchSpecification(criterias.get(0));
    if (spec == null) {
      return null;
    }

    final boolean isOrPredicate = orOpredicateFlag != null && (orOpredicateFlag.equals(SearchOperation.OR_PREDICATE_FLAG));
    for (int i = 1; i < criterias.size(); i++) {
      if (isOrPredicate) {
        spec = Specification.where(spec)
                .or(resource.getSearchSpecification(criterias.get(i)));
      } else {
        spec = Specification.where(spec)
                .and(resource.getSearchSpecification(criterias.get(i)));
      }
    }
    return spec;
  }

  /**
   * Checks that all linked RESTResources identified by their resId exist. Must be overridden if the
   * Resource has linked RESTResources
   *
   * @param Resource      item
   * @param BindingResult errors
   */
  public void validateLinkedResources(T item, BindingResult errors) {
  }

  /**
   * Override to set default data or clean properties
   *
   * @param item
   */
  protected void beforeValidate(T item) {
  }

  /**
   * Code executed after validation but before saving entity
   *
   * @param item
   */
  protected void beforeSave(T item) {
    this.setAuthenticatedUserProperties(item);
    item.updateTime();
  }

  protected void validate(T item) {

    final BindingResult errors = new BeanPropertyBindingResult(item, item.getClass().getName());

    this.validateConstraints(item, errors);
    this.validateItemSpecificRules(item, errors);

    if (errors.hasErrors()) {
      throw new SolidifyValidationException(new ValidationError(errors));
    }
  }

  /**
   * Validate Bean constraints
   *
   * @param Resource      item
   * @param BindingResult errors
   */
  protected void validateConstraints(T item, BindingResult errors) {
    this.validator.validate(item, errors);
  }

  /**
   * Override to add specific validation logic
   *
   * @param item
   * @param errors
   */
  protected void validateItemSpecificRules(T item, BindingResult errors) {
  }

  public T save(T item) {
    item.init();
    this.beforeValidate(item);
    this.validate(item);
    this.beforeSave(item);
    return this.localSave(item);
  }

  public abstract SolidifySpecification<T> getSpecification(T resource);
}
