package ch.unige.solidify.test.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;

import java.nio.charset.Charset;
import java.util.Random;
import java.util.concurrent.ThreadLocalRandom;

import org.junit.Test;

import ch.unige.solidify.util.StringCipherTools;

public class StringCipherToolsTest {

  private static final int MAX_PLAIN_TEXT_LENGTH = 200;
  private static final int MIN_PLAIN_TEXT_LENGTH = 1;
  private static final int NB_TRIES = 1_000;

  @Test
  public void cipherTestWithGeneratedSalt() {
    for (int i = 0; i < NB_TRIES; i++) {
      final String plainText = this.generateRandomPlainText();
      final String password = this.generateRandomPassword();
      final String cipherText = StringCipherTools.encrypt(plainText, password, 8);
      final String decryptedText = StringCipherTools.decrypt(cipherText, password, 8);
      assertEquals("Plain text (" + plainText + ") differs from decrypted text (" + decryptedText + ")", plainText, decryptedText);
      assertNotEquals("Plain text (" + plainText + ") equals cipher text (" + cipherText + ")", plainText, cipherText);
    }
  }

  @Test
  public void cipherTestWithPredefinedSalt() {
    for (int i = 0; i < NB_TRIES; i++) {
      final String plainText = this.generateRandomPlainText();
      final String password = this.generateRandomPassword();
      final String salt = this.generateRandomSalt();
      final String cipherText = StringCipherTools.encrypt(plainText, password, salt.getBytes());
      final String decryptedText = StringCipherTools.decrypt(cipherText, password, salt.getBytes().length);
      assertEquals("Plain text (" + plainText + ") differs from decrypted text (" + decryptedText + ")", plainText, decryptedText);
      assertNotEquals("Plain text (" + plainText + ") equals cipher text (" + cipherText + ")", plainText, cipherText);
    }
  }

  @Test
  public void undeterministCipher() {
    for (int i = 0; i < NB_TRIES; i++) {
      final String plainText = this.generateRandomPlainText();
      final String salt = this.generateRandomSalt();
      final String password = this.generateRandomPassword();
      final String cipherText1 = StringCipherTools.encrypt(plainText, password, salt.getBytes());
      final String cipherText2 = StringCipherTools.encrypt(plainText, password, salt.getBytes());
      assertNotEquals("Two encryption of the plain text (" + plainText + ") gives the same cipher", cipherText1, cipherText2);
    }
  }

  private String generateRandomPassword() {
    return this.generateRandomPlainText();
  }

  private String generateRandomPlainText() {
    final int plainTextLength = ThreadLocalRandom.current().nextInt(MIN_PLAIN_TEXT_LENGTH, MAX_PLAIN_TEXT_LENGTH + 1);
    final byte[] plainTextBytes = new byte[plainTextLength];
    new Random().nextBytes(plainTextBytes);
    return new String(plainTextBytes, Charset.forName("UTF-8"));
  }

  private String generateRandomSalt() {
    final byte[] saltBytes = new byte[8];
    new Random().nextBytes(saltBytes);
    return new String(saltBytes, Charset.forName("UTF-8"));
  }
}
